@@ -1,11 +1,16 @@
 from contextlib import contextmanager
+from datetime import datetime
 import select, json, re, psycopg2
 import psycopg2.extensions
 
-LISTEN_TIMEOUT_SECONDS = 60 # 1min
+def get_timespan(start):
+    delta = (datetime.now() - start)
+    return ((delta.days * 86400) + delta.seconds + (delta.microseconds / 1000000.0))
 
 class QueueManager(object):
 
+    LISTEN_TIMEOUT_INTERVAL_SECONDS = 1 # second
+
     def __init__(self,
                  dsn="", table_name="mq",
                  data_type="json",
@@ -75,16 +80,18 @@ create table %s (
     tag            varchar(31)     not null,
     content        varchar(%d),
     created_at     timestamp       not null default current_timestamp,
-    except_times   integer         default 0
+    except_times   integer         default 0,
+    schedule       timestamp
 );
 create index %s_tag_idx         on %s(tag);
 create index %s_created_at_idx  on %s(created_at);
-""" % (n, self.data_length, n, n, n, n,)
+create index %s_schedule        on %s(schedule);
+""" % (n, self.data_length, n, n, n, n, n, n)
         self.drop_table_sql = """
 drop table %s;
 """ % (n,)
         self.insert_sql = """
-insert into %s (tag, content) values ('%%s', '%%s');
+insert into %s (tag, content, schedule) values ('%%s', '%%s', %%s) returning id;
 """ % (n,)
         self.report_sql = """
 update %s set except_times = except_times + 1
@@ -93,16 +100,20 @@ update %s set except_times = except_times + 1
 """ % (n,)
         self.select_sql = """
 select * from %s
-  where case when tag = '%%s' then pg_try_advisory_lock(tableoid::int, id) else false end
+  where case
+    when (tag = '%%s' and (schedule is null or schedule <= current_timestamp))
+    then pg_try_advisory_lock(tableoid::int, id)
+    else false
+  end
   limit 1;
 """ % (n,)
         self.list_sql = """
 select * from %s
-  where case when tag = '%%s' then pg_try_advisory_lock(tableoid::int, id) else false end;
+  where case when (tag = '%%s'%%s) then pg_try_advisory_lock(tableoid::int, id) else false end;
 """ % (n,)
         self.count_sql = """
 select count(*) from %s
-  where case when tag = '%%s' then pg_try_advisory_lock(tableoid::int, id) else false end;
+  where case when (tag = '%%s'%%s) then pg_try_advisory_lock(tableoid::int, id) else false end;
 """ % (n,)
         self.cancel_sql = """
 delete from %s where id = %%s and pg_try_advisory_lock(tableoid::int, id)
@@ -133,13 +144,28 @@ listen %s;
         self.drop_table(other_sess)
         self.create_table(other_sess)
 
-    def enqueue(self, tag, data, other_sess = None):
+    def sanitize(self, string):
+        return string.replace("'", "''")
+
+    def check_tag(self, tag):
+        if "'" in tag:
+            raise ValueError("Invalid tag-name. invalid char \"'\" is in tag-name.")
+        return tag
+
+    def enqueue(self, tag, data, other_sess = None, schedule = None):
+        tag, data = (self.check_tag(tag), self.sanitize(self.serializer(data)), )
         with self.session(other_sess) as (conn, cur):
-            cur.execute((self.insert_sql + (self.notify_sql % (tag,))) % (tag, self.serializer(data),))
+            cur.execute(self.insert_sql % (
+                    tag, data,
+                    schedule.strftime('timestamp \'%Y-%m-%d %H:%M:%S\'') if schedule else 'NULL', ))
+            res = cur.fetchone()
+            cur.execute(self.notify_sql % (tag,))
             if conn: conn.commit()
+            return res[0] if res else None
 
     @contextmanager
     def dequeue_item(self, tag, other_sess = None):
+        tag = self.check_tag(tag)
         with self.session(other_sess) as (conn, cur):
             cur.execute(self.select_sql % (tag,))
             res = cur.fetchone()
@@ -168,6 +194,9 @@ listen %s;
             return
 
     def listen_item(self, tag, timeout=None):
+        tag         = self.check_tag(tag)
+        wait_start  = datetime.now()
+        interval    = self.LISTEN_TIMEOUT_INTERVAL_SECONDS
         while True:
             with self.session(None) as (conn, cur):
                 cur.execute(self.select_sql % (tag,))
@@ -177,16 +206,18 @@ listen %s;
                     if not ((0 < self.excepted_times_to_ignore) and
                             (self.excepted_times_to_ignore <= int(res[4]))):
                         yield res
+                        wait_start = datetime.now()
                     cur.execute(self.ack_sql % (res[0],))
                     conn.commit()
                     self.invoking_queue_id = None
                     continue
                 conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                 cur.execute((self.listen_sql % (tag,)))
-                if select.select([conn],[],[],(timeout or LISTEN_TIMEOUT_SECONDS)) == ([],[],[]):
-                    if timeout:
+                if select.select([conn],[],[],interval) == ([],[],[]):
+                    if timeout and (timeout <= get_timespan(wait_start)):
                         self.invoking_queue_id = None # to ignore error reporting.
                         yield None
+                        wait_start = datetime.now()
                     continue
                 conn.poll()
                 if conn.notifies:
@@ -198,6 +229,7 @@ listen %s;
                         if not ((0 < self.excepted_times_to_ignore) and
                                 (self.excepted_times_to_ignore <= int(res[4]))):
                             yield res
+                            wait_start = datetime.now()
                         cur.execute(self.ack_sql % (res[0],))
                         conn.commit()
                         self.invoking_queue_id = None
@@ -207,6 +239,7 @@ listen %s;
             yield (self.deserializer(d[2]) if d != None else None)
 
     def dequeue_immediate(self, tag, other_sess = None):
+        tag = self.check_tag(tag)
         with self.session(other_sess) as (conn, cur):
             cur.execute(self.select_sql % (tag,))
             res = cur.fetchone()
@@ -225,14 +258,18 @@ listen %s;
                 return res[0]
             return res[0] if res else False
 
-    def list(self, tag, other_sess = None):
+    def list(self, tag, other_sess = None, ignore_scheduled = True):
+        tag = self.check_tag(tag)
+        schedule = (" and (schedule is null or schedule <= current_timestamp)" if ignore_scheduled else "")
         with self.session(other_sess) as (conn, cur):
-            cur.execute(self.list_sql % (tag,))
+            cur.execute(self.list_sql % (tag, schedule, ))
             res = cur.fetchall()
             return res
 
-    def count(self, tag, other_sess = None):
+    def count(self, tag, other_sess = None, ignore_scheduled = True):
+        tag = self.check_tag(tag)
+        schedule = (" and (schedule is null or schedule <= current_timestamp)" if ignore_scheduled else "")
         with self.session(other_sess) as (conn, cur):
-            cur.execute(self.count_sql % (tag,))
+            cur.execute(self.count_sql % (tag, schedule, ))
             res = cur.fetchone()[0]
             return int(res)
